/*
 * at89-prog
 *
 * Copyright (C) 2010..2013  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <at89-prog.h>
#include <prog-io.h>
#include <spi.h>
#include <sched.h>

int verbose_level;
static unsigned long spi_freq = 11059200 / 300;

static unsigned int rom_size = 8 * 1024;

static const char *device = NULL;
static const char *image = NULL;

struct chip {
	const char *model;
	unsigned int vendor_id;
	unsigned int device_id;
	unsigned int page_size;
	unsigned int num_pages;
};

#define VENDOR_ATMEL	0x1e

#define CHIP(_model, _vendor_id, _device_id, _page_size, _num_pages) { \
		.model = (_model), \
		.vendor_id = (_vendor_id), \
		.device_id = (_device_id), \
		.page_size = (_page_size), \
		.num_pages = (_num_pages), \
	}

static struct chip chips[] = {
	CHIP("AT89S52", VENDOR_ATMEL, 0x5206, 256, 32),
	CHIP("AT89S8253", VENDOR_ATMEL, 0x5206, 256, 32),
	{ NULL },
};

static void show_info(void)
{
	fprintf(stderr, "at89-prog "PACKAGE_VERSION"\n"
			"Copyright (C) 2010  "
			"Krzysztof Mazur <krzysiek@podlesie.net>\n\n");
}

static void show_usage(char *name)
{
	show_info();
	fprintf(stderr,	"usage: %s [options] (file)\n"
			"Options:\n"
			"  -d, --device		use ppdev device\n"
			"  -h, --help		show this help\n"
			"  -q, --quiet		quiet mode\n"
			"  -v, --verbose		verbose mode\n"
			"  -V, --version		show version\n",
			(name == NULL) ? "at89-prog" : name);
	exit(EXIT_SUCCESS);
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"help", 0, 0, 'h'},
		{"quiet", 0, 0, 'q'},
		{"verbose", 0, 0, 'v'},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};

	while (1) {
		c = getopt_long(argc, args, "hqvV", long_options,
				&option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'h':
			show_usage(args[0]);
			break;
		case 'q':
			verbose_level--;
			break;
		case 'v':
			verbose_level++;
			break;
		case 'V':
			exit(EXIT_SUCCESS);
			break;
		default:
			fprintf(stderr, "Unknown option.\n");
			exit(EXIT_FAILURE);
		}
	}

	if (optind < argc) {
		image = args[optind++];
		if (optind < argc)
			fprintf(stderr, "warning: unknown parameters detected."
					"\n");
	}
}

static unsigned char image_buf[65536];
static unsigned int image_size;

void image_write(unsigned int addr, uint8_t data)
{
	if (addr >= 65536)
		abort();
	if (addr >= rom_size) {
		fprintf(stderr, "error: address %d is out of ROM\n", addr);
		exit(EXIT_FAILURE);
	}

	image_buf[addr] = data;
	if (addr + 1 > image_size)
		image_size = addr + 1;
}

int hex_to_int(char x)
{
	if (x >= '0' && x <= '9')
		return x - '0';
	if (x >= 'A' && x <= 'F')
		return x - 'A' + 10;
	if (x >= 'a' && x <= 'f')
		return x - 'a' + 10;
	return -1;
}

int parse_hex_byte(const char *x)
{
	int lo = hex_to_int(x[1]);
	int hi = hex_to_int(x[0]);

	if (lo < 0 || hi < 0)
		return -1;

	return (hi << 4) | lo;
}

static void load_image(const char *path)
{
	FILE *f;
	char *line = NULL;
	size_t line_size = 0;
	ssize_t len;
	int data_len;
	int cmd;
	int addr_lo;
	int addr_hi;
	unsigned long addr;
	unsigned char x;
	unsigned int i;

	f = fopen(path, "r");
	if (f == NULL) {
		perror("fopen");
		exit(EXIT_FAILURE);
	}

	while (!feof(f)) {
		len = getline(&line, &line_size, f);

		if (len < 2)
			break;

		if (line[0] != ':') {
			fprintf(stderr, "corrupted image\n");
			break;
		}

		data_len = parse_hex_byte(line + 1);
		if (data_len < 0) {
			fprintf(stderr, "corrupted image\n");
			break;
		}

		if (2 * data_len + 2 + 4 * 2 > len) {
			fprintf(stderr, "corrupted image\n");
			break;
		}

		cmd = parse_hex_byte(line + 7);
		if (cmd < 0) {
			fprintf(stderr, "corrupted image\n");
			break;
		}

		if (cmd == 1)
			break;

		if (cmd != 0) {
			fprintf(stderr, "unsupported IHEX command %d\n", cmd);
			break;
		}

		addr_hi = parse_hex_byte(line + 3);
		addr_lo = parse_hex_byte(line + 5);
		if (addr_hi < 0 || addr_lo < 0) {
			fprintf(stderr, "corrupted image: invalid addr\n");
			break;
		}

		addr = (addr_hi << 8) | addr_lo;

		for (i = 0; i < data_len; i++) {
			x = parse_hex_byte(line + 9 + 2 * i);
			image_write(addr++, x);
		}
	}
}

static void dump_page(uint8_t *a, uint8_t *b) __attribute__((unused));

static void dump_page(uint8_t *a, uint8_t *b)
{
	unsigned int i;

	for (i = 0; i < 64; i++)
		fprintf(stderr, "x[%d]	%02x %02x\n", i, a[i], b[i]);
}
	
static struct chip *find_chip(u8 *sig, unsigned int len)
{
	u8 vendor_id;
	u16 device_id;
	struct chip *c;

	if (len != 3)
		return NULL;
	vendor_id = sig[0];
	device_id = ((u16) sig[1] << 8) | sig[2];
	for (c = &chips[0]; c->model; c++)
		if (c->vendor_id == vendor_id && c->device_id == device_id)
			return c;
	return NULL;
}

int main(int argc, char **args)
{
	struct chip *c;
	unsigned int i;
	unsigned int page;
	unsigned int retry = 10;
	unsigned int to_write;
	u8 *page_buf;
	u8 *page_vbuf;
	int ret;
	u8 sig[3];

	parse_command_line(argc, args);
	if (image == NULL)
		show_usage(args[0]);

	load_image(image);

	fprintf(stderr, "Loaded %d bytes image\n", image_size);

	ppdev_init(device);
	spi_init(spi_freq);

	usleep(10000);

	fprintf(stderr, "Enabling programming mode...");
	ret = prog_enable();
	if (ret) {
		fprintf(stderr, "FAILED\n");
		fprintf(stderr, "Cannot enable programming mode.\n");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "OK\n");

	fprintf(stderr, "Reading AT89S52 chip signature...");
	ret = read_signature(sig, 3);
	if (ret != 3) {
		fprintf(stderr, "FAILED\n");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "OK (0x%02x 0x%02x 0x%02x)\n", sig[0], sig[1], sig[2]);

	c = find_chip(sig, 3);
	if (c == NULL) {
		fprintf(stderr, "error: unknown chip detected\n");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "Found %s, ROM %ld kB (page size %d)\n", c->model,
			(unsigned long) c->page_size * c->num_pages >> 10,
			c->page_size);

	page_buf = malloc(c->page_size);
	page_vbuf = malloc(c->page_size);
	if (page_buf == NULL || page_vbuf == NULL) {
		fprintf(stderr, "Cannot allocate page buffer\n");
		exit(EXIT_FAILURE);
	}

	fprintf(stderr, "Erasing chip...");
	chip_erase();
	usleep(100000);
	fprintf(stderr, "OK\n");

	fprintf(stderr, "Enabling programming mode...");
	ret = prog_enable();
	if (ret) {
		fprintf(stderr, "FAILED\n");
		fprintf(stderr, "Cannot enable programming mode.\n");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "OK\n");

	fprintf(stderr, "Checking first page...");
	read_page(0, page_buf, c->page_size);
	for (i = 0; i < c->page_size; i++)
		if (page_buf[i] != 0xff) {
			fprintf(stderr, "error\n");
			fprintf(stderr, "Page not erased\n");
			exit(EXIT_FAILURE);
		}
	fprintf(stderr, "OK\n");

	to_write = (image_size + c->page_size) / c->page_size;
	if (to_write >= c->num_pages) {
		fprintf(stderr, "image too big: %d pages, only %d available\n",
				to_write, c->num_pages);
		exit(EXIT_FAILURE);
	}

	for (page = 0; page < to_write; ) {
		fprintf(stderr, "programming page %d/%d\n", page, to_write);
		memcpy(page_buf, image_buf + page * c->page_size, c->page_size);
		write_page(page, page_buf, c->page_size);
		usleep(5000);
		read_page(page, page_vbuf, c->page_size);
		if (memcmp(page_vbuf, page_buf, c->page_size)) {
			fprintf(stderr, "page %d: write error\n", page);
			retry--;
			if (!retry) {
				fprintf(stderr, "retry limit reached\n");
				exit(EXIT_FAILURE);
			}
			continue;
		}
		page++;
	}

	fprintf(stderr, "Deasserting reset...\n");
	pin_set(PIN_RST, 0);
	
	return 0;
}
