/*
 * at89-prog - ppdev.c
 *
 * 	I/O interface on ppdev
 *
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <linux/ppdev.h>

#include <prog-io.h>

enum {
	PORT_DATA,
	PORT_CONTROL,
	PORT_STATUS,
};

struct pin {
	int port;
	int negated;			/* true if signal is negated */
	int bit;			/* */
};

#define DATA_PIN(bit)		{ PORT_DATA, 0, (bit) }
#define DATA_NPIN(bit)		{ PORT_DATA, 1, (bit) }
#define CTRL_PIN(bit)		{ PORT_CONTROL, 0, (bit) }
#define CTRL_NPIN(bit)		{ PORT_CONTROL, 1, (bit) }
#define STAT_PIN(bit)		{ PORT_STATUS, 0, (bit) }
#define STAT_NPIN(bit)		{ PORT_STATUS, 1, (bit) }

struct interface {
	const char *name;
	struct pin pins[PIN_COUNT];
};

static struct interface interfaces[] = {
	/* name,    RST          SCK          MOSI         MISO */
	{ "AT89", { CTRL_NPIN(3), CTRL_NPIN(0), DATA_PIN(0), STAT_PIN(6) } }
};

static struct interface *iface = &interfaces[0];

static int ppdev_fd = -1;

static unsigned char data = 0xff;
static unsigned char control = 0xdf;
static unsigned char status = 0xff;
static int data_dir = -1;

static void data_write(void)
{
	int ret;

	if (data_dir != 0) {
		data_dir = 0;

		ret = ioctl(ppdev_fd, PPDATADIR, &data_dir);
		if (ret)
			perror("ioctl: PPDATADIR");
	}

	ret = ioctl(ppdev_fd, PPWDATA, &data);
	if (ret)
		perror("ppdev: PPWDATA: ");
}

static void data_read(void)
{
	int ret;

	if (data_dir != 1) {
		data_dir = 1;
		ret = ioctl(ppdev_fd, PPDATADIR, &data_dir);
		if (ret)
			perror("ioctl: PPDATADIR");
	}

	ret = ioctl(ppdev_fd, PPRDATA, &data);
	if (ret)
		perror("ppdev: PPRDATA: ");
}

static void control_write(void)
{
	int ret;

	ret = ioctl(ppdev_fd, PPWCONTROL, &control);
	if (ret)
		perror("ppdev: PPWCONTROL: ");
}

static void control_read(void)
{
	int ret;

	ret = ioctl(ppdev_fd, PPRCONTROL, &control);
	if (ret)
		perror("ppdev: PPRCONTROL: ");
}

static void status_read(void)
{
	int ret;

	ret = ioctl(ppdev_fd, PPRSTATUS, &status);
	if (ret)
		perror("ppdev: PPRSTATUS: ");
}

static unsigned char bit_copy(unsigned char x, int bit, int negated, int value)
{
	if (!!value ^ !!negated) {
		x |= 1 << bit;
	} else {
		x &= ~(1 << bit);
	}
	return x;
}

static int read_bit(unsigned char x, int bit, int negated)
{
	if (negated)
		return !(x & (1 << bit));
	return !!(x & (1 << bit));
}

void __pin_set(struct pin *pin, int value)
{
	if (pin->port == PORT_DATA) {
		data = bit_copy(data, pin->bit, pin->negated, value);
		data_write();
	} else if (pin->port == PORT_CONTROL) {
		control = bit_copy(control, pin->bit, pin->negated, value);
		control_write();
	}
}

int __pin_get(struct pin *pin)
{
	int ret = 0;

	switch (pin->port) {
	case PORT_DATA:
		data_read();
		ret = read_bit(data, pin->bit, pin->negated);
		break;
	case PORT_CONTROL:
		control_read();
		ret = read_bit(control, pin->bit, pin->negated);
		break;
	case PORT_STATUS:
		status_read();
		ret = read_bit(status, pin->bit, pin->negated);
		break;
	}
	return ret;
}

void pin_set(int pin, int value)
{
	__pin_set(&iface->pins[pin], value);
}

int pin_get(int pin)
{
	return __pin_get(&iface->pins[pin]);
}
	
static void ppdev_exit(void)
{
	int ret;

	ret = ioctl(ppdev_fd, PPRELEASE);
	if (ret)
		perror("ioctl: PPRELEASE");

	ret = close(ppdev_fd);
	if (ret)
		perror("close");
}

int ppdev_init(const char *path)
{
	int ret;

	if (path == NULL)
		path = "/dev/parport0";

	ppdev_fd = open(path, O_RDWR);
	if (ppdev_fd == -1) {
		perror("open");
		return 1;
	}

	ret = ioctl(ppdev_fd, PPCLAIM);
	if (ret) {
		perror("ioctl: PPCLAIM");
		return 1;
	}

	data = 0xff;
	data_write();
	status = 0xff;

	pin_set(PIN_SCK, 0);
	pin_set(PIN_RST, 0);
	usleep(100000);
	pin_set(PIN_RST, 1);
	usleep(100000);

	atexit(ppdev_exit);
	return 0;
}
